#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_snow
----------------------------------

Tests for `envphys.snow` module.
"""

import unittest

from smrf.envphys import snow
import numpy as np

class TestNASDE(unittest.TestCase):
    def testMarks2017(self):
        """
        Tests the New Snow Density Model : marks2017
        """

        T = np.array([-10,-5,0,1])
        PP = np.array([100,100,100,100])
        s = snow.marks2017(T,PP)
        r = [79,115,250,0]

        for i,rho in enumerate(s['rho_s']):
            msg = "NASDE Model Failed: marks2017({0},{3}) == {1}\nExpected {2}".format(T[i],rho,r[i],PP[i])
            self.assertAlmostEqual(rho, r[i], places = 0, msg= msg)

    def testSusong1999(self):
        """
        Tests the New Snow Density Model : susong1999
        """

        T = np.array([-5,-3,0])
        PP = np.array([1,1,1])

        s = snow.susong1999(T, PP)
        r = [100,150,250]
        for i,rho in enumerate(s['rho_s']):
            msg = "NASDE Model Failed: susong1999({0},{3}) == {1}\nExpected {2}".format(T[i],rho,r[i],PP[i])
            self.assertAlmostEqual(rho, r[i],places = 0, msg= msg)


if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



# !/usr/bin/env python

import argparse
from . output import print_config_report, generate_config, print_recipe_summary, print_details
from . tools import get_user_config, check_config
import os
import sys
from config import MasterConfig
from . utilities import pcfg

def main():

    parser = argparse.ArgumentParser(description="Examine and auto populate"
                                                 " ini files with a master file"
                                                 " comparison.")
    parser.add_argument('-f', '--config_file', dest='config_file', type=str,
                        help='Path to a config file that needs checking')

    parser.add_argument('--master', '-c', metavar='MF', type=str,
                        help='Path to a config file that used to check against')

    parser.add_argument('--module', '-m', metavar='M', type=str,
                        help="Module name with an attribute __CoreConfig__ that"
                             " is a path to a master config file for checking"
                             " against")

    parser.add_argument('-w', '--write', dest='write', action='store_true',
                        help="Determines whether to write out the file with all"
                             " the defaults")

    parser.add_argument('-r', '--recipes', dest='recipes', action='store_true',
                        help="Prints out the recipe summary")


    parser.add_argument('--details', '-d', type=str, nargs='+', help="Provide
                        "section item and value for details regarding them")
    args = parser.parse_args()

    # Module not provided, or master config
    if args.module == None and args.master == None:
        print("ERROR: Please provide either a module or a path to a master
             " config, or ask for details on config entries")
        sys.exit()


    # Normal operation
    else:
        # Requesting details for a section and item
        if args.details != None:
            if len(args.details) == 1:
                print("Providing details for section {0}..."
                "".format(args.details[0]))

            elif len(args.details) == 2:
                print("Providing details for section {0} and item {1}...
                      "".format(args.details[0], args.details[1]))

            else:
                print("Details option can at most recieve section and item ")
                sys.exit()

            mcfg = MasterConfig(path=args.master, module=args.module)
            print_details(args.details, mcfg.cfg)

        else:
            f = os.path.abspath(args.config_file)
            ucfg = get_user_config(f, master_files=args.master,
                                      module=args.module)
            pcfg(ucfg.cfg)
            warnings, errors = check_config(ucfg)

            print_config_report(warnings, errors)

            if args.recipes:
                print_recipe_summary(ucfg.recipes)

            if args.write:
                out_f = './{0}_full.ini'
                ''.format(os.path.basename(f).split('.')[0])

                print("Writing complete config file with all recipes and "
                      " necessary defaults...")
                print('{0}'.format(out_f))

                generate_config(ucfg, out_f, section_titles=ucfg.mcfg.titles,
                                             cli=True)


if __name__ == '__main__':
    main()



import os
from . config import UserConfig, MasterConfig
from . utilities import mk_lst, get_checkers

def check_config(config_obj):
            """
            looks at the users provided config file and checks it to a master
            config file looking at correctness and missing info.

            Returns:
                tuple:
                - **warnings** - Returns a list of string messages that are
                                 consider non-critical issues with config file.
                - **errors** - Returns a list of string messages that are
                               consider critical issues with the config file.
            """

            msg = "{: <20} {: <30} {: <60}"
            errors = []
            warnings = []

            master = config_obj.mcfg.cfg
            cfg = config_obj.cfg

            standard_funcs = get_checkers()

            # Add any checker modules if provided
            if config_obj.mcfg.checker_module != None:
                funcs = standard_funcs.update(
                        get_checkers(module=config_obj.mcfg.checker_module))

            # Compare user config file to our master config
            for section, configured in cfg.items():

                # In the section check the values and options
                for item, value in configured.items():
                    litem = item.lower()

                    # Is the item known as a configurable item?
                    if litem not in master[section].keys():
                        wrn = "Not a registered option."
                        warnings.append(msg.format(section, item, wrn))
                    else:
                        # Did the user provide a list value or single value
                        val_lst = mk_lst(value)

                        for v in val_lst:
                            if v != None:
                                # Do we have an idea of what to expect?
                                options_type = master[section][item].type
                                for name, fn in standard_funcs.items():

                                    # Check for type checkers
                                    if options_type == name.lower():
                                        b = fn(value=v, config=config_obj)
                                        issue = b.check()
                                        if issue != None:
                                            full_msg = msg.format(section,
                                                                  item,
                                                                  issue)
                                            if b.msg_level == 'error':
                                                errors.append(full_msg)
                                            elif b.msg_level == 'warning':
                                                warnings.append(full_msg)
                                            break
                                    else:
                                        issue = None
            return warnings, errors


def cast_all_variables(config_obj, mcfg_obj):
    """
    Cast all values into the appropiate type using checkers, other_types
    and the master config.

    Args:
        config_obj: The object of the user config from class UserConfig
        mcfg_obj: The object used for manage the master config from
                  class MasterConfig
        other_types: User provided list to add any custom types

    Returns:
        ucfg: The users config dictionary containing the correct value types

    """

    ucfg = config_obj.cfg
    mcfg = mcfg_obj.cfg
    all_checks = get_checkers()

    # add in other type casting
    if mcfg_obj.checker_module != None:
        all_checks.update(get_checkers(module=mcfg_obj.checker_module))

    # Cast all variables
    for s in ucfg.keys():
        for i in ucfg[s].keys():
            values = []

            # Ensure it is something we can check
            if i in mcfg[s].keys():
                type_value = (mcfg[s][i].type).lower()
                if type_value.lower() not in all_checks.keys():
                    raise ValueError("Type {0} is undefined and has no"
                                    " checker associated"
                                    "".format(type_value))

                for z, v in enumerate(mk_lst(ucfg[s][i])):
                    option_found = False
                    # Use all the checks available to cast
                    for name, fn in all_checks.items():
                        if  type_value == name.lower():
                            b = fn(value=v, config=config_obj)
                            if v in [None, 'none', 'None']:
                                values.append(None)
                            else:
                                values.append(b.cast())
                            option_found = True
                            break

                    if not option_found:
                        raise ValueError("Unknown type_value prescribed."
                                         " ----> {0}".format(type_value))
            else:
                values = ucfg[s][i]

            ucfg[s][i] = mk_lst(values, unlst=True)

    config_obj.cfg = ucfg
    return config_obj


def get_user_config(config_file, master_files=None, module=None,
                    mcfg=None):
    """
    Returns the users config as the object UserConfig.

    Args:
        config_file: real path to existing config file
        master_file: real path to a Core Config file
        module: a module with a string attribute __CoreConfig__ which is the
                path to a CoreConfig
        mcfg: the master config object after it has been read in.

    Returns:
        ucfg: Users config as an object
    """

    if module == None and master_files == None:
        raise IOError("ERROR: Please provide either a module or a path to a"
                      "master config")
        sys.exit()

    if os.path.isfile(config_file):

        master_files = mk_lst(master_files)

        mcfg = MasterConfig(path=master_files, module=module)

        ucfg = UserConfig(config_file, mcfg=mcfg)
        ucfg.apply_recipes()
        ucfg = cast_all_variables(ucfg, mcfg)


    else:
        raise IOError("Config file path {0} doesn't exist."
                      "".format(config_file))

    return ucfg


def config_documentation(out_f, paths=None, module=None,
                        section_link_dict=None):
    """
    Auto documents the core config file. Outputs to a file which is can then be
    used for documentation. Specifically formulated for sphinx
    """

    if paths == None and module == None:
        raise ValueError("inicheck function config_documentation args paths or"
                        " module must be specified!")


    master = MasterConfig(path=paths, module=module)
    mcfg = master.cfg

    #RST header
    config_doc = "Configuration File Reference\n"
    config_doc += "============================\n"
    config_doc += "\nFor configuration file syntax information please visit"
    config_doc += " http://inicheck.readthedocs.io/en/latest/"

    for section in mcfg.keys():
        # Section header
        config_doc += " \n"
        config_doc += "{0}\n".format(section)
        config_doc += "-" * len(section) + '\n'

        # Allow for custom info in a description about the section, useful
        # for linking modules for auto documenting
        if section in section_link_dict.keys():
            intro = section_link_dict[section]
        else:
            intro = ''

        config_doc += intro
        config_doc += "\n"

        # Auto document config file according to master config contents
        for item,v in sorted(mcfg[section].items()):
            # Check for attributes that are lists
            for att in ['default', 'options']:
                z = getattr(v,att)
                if type(z) == list:
                    combo = ' '
                    doc_s = combo.join([str(s) for s in z])
                    setattr(v,att,doc_s)

            # Bold item with definition
            config_doc += "| **{0}**\n".format(item)

            # Add the item description
            config_doc += "| \t{0}\n".format(v.description)

            # Default
            config_doc += "| \t\t*Default: {0}*\n".format(v.default)

            # Add expected type
            config_doc += "| \t\t*Type: {0}*\n".format(v.type)

            # Print options should they be available
            if v.options:
                config_doc += "| \t\t*Options:*\n *{0}*\n".format(v.options)

            config_doc += "| \n"

            config_doc += "\n"

    path = os.path.abspath(out_f)
    if not os.path.isfile(path):
        raise IOError("Config Documentation output file does not exist."
                      "\n{0}".format(out_f))

    print("Writing auto documentation for config file to:\n{0}".format(path))

    with open(path,'w+') as f:
        f.writelines(config_doc)
    f.close()



# -*- coding: utf-8 -*-
import os
__author__ = """Micah Johnson"""
__email__ = 'micah.johnson150@gmail.com'
__version__ = '0.2.3'

# Keywords for conditional statements in a master config file
__trigger_keywords__ = ['trigger', 'triggers', 'condition']

# Key words expected in the section name of a recipe
__recipe_keywords__ = ['recipe']

__core_config__ = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                  '../examples/CoreConfig.ini'))



'''
Functions for checking valus in a config file and producing errors and warnings
'''
import os
from pandas import to_datetime


class GenericCheck(object):
    def __init__(self, **kwargs):
        if 'value' not in kwargs.keys():
            raise ValueError("Must provided at least keyword value to "
                             "Checkers.")
        if 'config' not in kwargs.keys():
            raise ValueError("Must provided at least keyword config to"
                             "Checkers.")

        self.message = None

        self.msg_level = 'warning'
        self.value = kwargs["value"]

        self.config = kwargs["config"]

        if not self.msg_level.lower() in ['warning', 'error']:
            raise ValueError("msg_level = {0} not allowed."
                             "".format(self.msg_level))

    def is_valid(self):
        """
        Abstract function for defining how a value is checked for validity

        Args:
            value: Value that is going to be check for validity
        Returns:
            **(boolean, msg)**: True or False whether value valid and
                                correspond error message if it is not
        """
        pass

    def check(self):
        """
        Function that is ran by the checking function of the user config to
        return useful message to instruct user

        Args:
            value: value to be check for validity
        Returns:
            msg: None is the entry is valid, else returns self.message
        """

        msg = None
        valid = self.is_valid()
        if not valid:
            msg = self.message
        return msg


class CheckType(GenericCheck):
    """
    Base class for type checking whether a value of the right type.
    """

    def __init__(self, **kwargs):
        super(CheckType, self).__init__(**kwargs)
        self.type = 'string'
        # Function used for casting to types
        self.type_func = None

    def is_valid(self):
        """
        Checks for type validity
        """
        msg = None
        self.msg_level = 'error'
        if self.type not in str(type(self.value)):
                valid = False
                msg = "Expecting {0} received {1}".format(self.type,
                                                          str(self.value))
        else:
            valid = True
        return valid, msg

    def cast(self):
        return self.type_func(self.value)


class CheckDatetime(CheckType):
    """
    Check values that are declared as type datetime.
    """

    def __init__(self, **kwargs):
        super(CheckDatetime, self).__init__(**kwargs)
        self.type_func = to_datetime

    def is_valid(self):
        """
        Checks for datetime
        """
        msg = None
        try:
            self.cast()
            return True, msg
        except:
            msg = "Not in datetime format"
            return False, msg


class CheckFloat(CheckType):
    """
    Float checking whether a value of the right type.
    """

    def __init__(self, **kwargs):

        super(CheckFloat, self).__init__(**kwargs)
        self.type_func = float
        self.type = 'float'


class CheckInt(CheckType):
    """
    integer checking whether a value of the right type.
    """

    def __init__(self, **kwargs):

        super(CheckInt, self).__init__(**kwargs)
        self.type_func = float
        self.type = 'float'

    def cast(self):
        return(self.type_func(self.value))


class CheckBool(CheckType):
    """
    Boolean checking whether a value of the right type.
    """

    def __init__(self, **kwargs):

        super(CheckBool, self).__init__(**kwargs)
        self.type_func = bool
        self.type = 'bool'

    def cast(self):
        if self.value.lower() in ['y', 'yes', 'true']:
            self.value = True
        elif self.value.lower() in ['n', 'no', 'false']:
            self.value = False

        return self.value


class CheckString(CheckType):
    """
    Float checking whether a value of the right type.
    """

    def __init__(self, **kwargs):

        super(CheckString, self).__init__(**kwargs)
        self.type_func = str
        self.type = 'string'


class CheckPath(CheckType):
    """
    Checks whether a Path exists.
    """

    def __init__(self, **kwargs):
        super(CheckPath, self).__init__(**kwargs)
        self.root_loc = self.config.filename

        self.dir_path = False

        # Path should alsways be absolute or relative to the config file path
        if self.value != None and self.root_loc != None:
            if not os.path.isabs(self.value):
                p = os.path.abspath(
                            os.path.expanduser(os.path.dirname(self.root_loc)))
                self.value = os.path.join(p, self.value)

    def is_valid(self):
        """
        Checks for existing filename
        """
        self.msg_level = 'warning'

        if self.dir_path:
            exists = os.path.isdir(self.value)
        else:
            exists = os.path.isfile(self.value)

        return exists

    def cast(self):
        return self.value


class CheckDirectory(CheckPath):
    """
    Checks whether a directory exists.
    """

    def __init__(self, **kwargs):
        super(CheckDirectory, self).__init__(**kwargs)
        self.dir_path = True
        self.message = "Directory does not exist."


class CheckFilename(CheckPath):
    """
    Checks whether a directory exists.
    """

    def __init__(self, **kwargs):
        super(CheckFilename, self).__init__(**kwargs)
        self.message = "File does not exist."


class CheckCriticalFilename(CheckFilename):
    """
    Checks whether a critical file exists.
    """

    def __init__(self, **kwargs):
        super(CheckCriticalFilename, self).__init__(**kwargs)
        self.msg_level = 'error'


class CheckCriticalDirectory(CheckDirectory):
    """
    Checks whether a critical directory exists.
    """

    def __init__(self, **kwargs):
        super(CheckCriticalDirectory, self).__init__(**kwargs)
        self.msg_level = 'error'



import inspect
import sys
import os


def mk_lst(values, unlst=False):
    """
    while iterating through several type of lists and items it is convenient to
    assume that we recieve a list, use this function to accomplish this. It also
    convenient to be able to return the original type after were done with it.
    """

    if type(values) != list:
        values = [values]
    else:
        if unlst:
            if len(values) == 1:
                values = values[0]

    return values

def remove_chars(orig_str, char_str, replace_str=None):
    """
    Removes all character in orig_str that are in char_str

    Args:
        orig_str: Original string needing cleaning
        char_str: String of characters to be removed
        replace_str: replace values with a character if requested

    Returns:
        string: orig_str with out any characters in char_str
    """
    if replace_str == None:
        clean_str = [c for c in orig_str if c not in char_str]
    else:
        clean_str = [c if c not in char_str else replace_str for c in orig_str]

    return "".join(clean_str)


def get_checkers(module='inicheck.checkers', keywords=[]):
    """
    Args:
        module: The module to search for the classes

    Returns a dictionary of the classes available for checking config entries
    """
    keywords.append('check')

    funcs = inspect.getmembers(sys.modules[module], inspect.isclass)
    func_dict = {}
    for name, fn in funcs:
        checker_found = False
        k = name.lower()
        # Remove any keywords
        for w in keywords:
            if w in k:
                k = k.replace(w, '')
                checker_found = True
                break

        if checker_found:
            func_dict[k] = fn

    return func_dict


def get_relative_to_cfg(path, user_cfg_path):
    if not os.path.isabs(path):
        path = os.path.abspath(os.path.join(os.path.dirname(user_cfg_path), path))
    return path


def pcfg(cfg):
    """
    prints out the config file to the prompt with a nice stagger Look for
    readability

    Args:
        cfg: dict of dict in a heirarcy of  {section, {items, values}}
    """

    for sec in cfg.keys():
        print(repr(sec))
        try:
            for item in cfg[sec].keys():
                print('\t' + item)
                values = cfg[sec][item]

                if type(cfg[sec][item]) == list:
                    out = ", ".join(cfg[sec][item])
                else:
                    out = cfg[sec][item]
                print('\t\t' + repr(out))
        except:
            print('\t recipe')


def pmcfg(cfg):
    """
    prints out the core config file to the prompt with a nice stagger Look for
    readability

    Args:
        cfg: dict of dict in a heirarcy of  {section, {items, values}}
    """

    for sec in cfg.keys():
        print(repr(sec))
        for item in cfg[sec].keys():
            print('\t' + item)
            obj = cfg[sec][item]
            for att in ['type', 'options', 'default', 'description']:
                value = getattr(obj, att)
                print('\t\t' + att)

                if type(value) == list:
                    out = ", ".join(value)
                else:
                    out = value
                print('\t\t\t' + repr(out))



from datetime import date
import os


def generate_config(config_obj, fname, package_header=None, cli=False):
    """
    Generates a list of strings to be written and then writes them in the ini
    file

    Args:
        config: Config file dictionary created by
                 :func:`~inicheck.config.UserConfig'.
        fname: String path to the output location for the new config file.
        package_header: This is string that enables the user to customize
                         config files with their own titles. Creating a string
                         at the top that says "Configuration File for
                         <package_header>"

        cli: Boolean value that adds the line "file generated using
                   inicheck.cli, Default = False

    Returns:
        None
    """
    header_len = 80
    pg_sep = '#' * header_len

    # Header surround each commented titles in the ini file
    section_header = pg_sep + '\n' + ('# {0}\n') + pg_sep

    # Construct the section strings
    config_str = ""
    config_str += pg_sep

    # File header with specific package option
    if config_obj.mcfg.header != None:
        header = config_obj.mcfg.header.split('\n')
        for line in header:
            config_str += ('\n# ' + line)
    else:
        config_str += "\n# Configuration File "

    # Add in the date generated
    config_str += "\n#\n# Date generated: {0}".format(date.today())

    # Generated with inicheck
    if cli:
        config_str += "\n#\n# Generated using: inicheck <filename> -w"

    config_str += "\n#\n# For more inicheck help see:" + \
                  "\n# http://inicheck.readthedocs.io/en/latest/\n"

    config = config_obj.cfg
    mcfg = config_obj.mcfg.cfg

    # Check to see if section titles were provided
    has_section_titles = hasattr(config_obj.mcfg,'titles')

    # Generate the string for the file, creating them in order.
    for section in mcfg.keys():
        if section in config.keys():
            config_str += '\n' * 2

            # Add a section header
            s_hdr = pg_sep
            if has_section_titles:
                if section in config_obj.mcfg.titles.keys():
                    # Add the header
                    s_hdr = section_header.format(config_obj.mcfg.titles[section])

            else:
                config_str += s_hdr

            config_str += s_hdr
            config_str += '\n'
            config_str += '\n[{0}]\n'.format(section)

            # Add section items and values
            for k in config[section].keys():
                v = config[section][k]
                if type(v) == list:
                    astr = ", ".join(str(c.strip()) for c in v)
                else:
                    astr = str(v)
                config_str += "{0:<30} {1:<10}\n".format((k+':'), astr)

    # Write out the string generated
    with open(os.path.abspath(fname), 'w') as f:
        f.writelines(config_str)
        f.close()


def print_config_report(warnings, errors, logger=None):
    """
    Pass in the list of string messages generated by check_config file.
    print out in a pretty format the issues

    Args:
        warnings - List of non-critical messages returned from
                   :func:`~utilities.check_config'.
        errors - List of critical messages returned from
                 :func:`~utilities.check_config'.
        logger - pass in the logger function being used. If no logger is
                 provided, print is used. Default = None

    Returns:
        None
    """

    msg = "{: <20} {: <25} {: <60}"

    # Check to see if user wants the logger or stdout
    if logger != None:
        out = logger.info
    else:
        out = print_out


    msg_len = 90
    out(" ")
    out(" ")
    out("Configuration File Status Report:")
    header = "="*msg_len
    out(header)
    any_warnings = False
    any_errors = False

    # Output warnings
    if len(warnings) > 0:
        any_warnings = True
        out("WARNINGS:")
        out(" ")
        out(msg.format(" Section", "Item", "Message"))
        out("-"*msg_len)
        for w in warnings:
            out(w)
        out(" ")
        out(" ")

    # Output errors
    if len(errors) > 0:
        any_errors = True
        out("ERRORS:")
        out(" ")
        out(msg.format("Section", "Item", "Message"))
        out("-"*msg_len)
        for e in errors:
            out(e)
        out(" ")
        out(" ")

    if not any_errors and not any_warnings:
        out("No errors or warnings were reported with the config file.\n")


def print_out(out_str):
    """
    wrapper for print so we can use either a logger or a stdout
    """
    print(out_str)


def print_recipe_summary(lst_recipes):
    """
    Prints out the recipes found and how they are interpretted
    """
    # len of recipe separators
    msg_len = 80
    header = "=" * msg_len
    recipe_hdr = "-" * msg_len
    r_msg = "\n{: <20}\n"+recipe_hdr
    cfg_msg = "\t\t{: <20} {: <20} {: <20}"

    msg = "\t\t{: <20} {: <25}"

    print('\n\n')
    print("Below are the recipes applied to the config file:")
    print("Recipes Summary:")
    print(header)

    for r in lst_recipes:
        print(r_msg.format(r.name))
        print("\tConditionals:")
        for n, t in r.triggers.items():
            for i, c in enumerate(t.conditions):
                if i == 0:
                    print(msg.format(n, c))
                else:
                    print(msg.format("", c))

        print_cfg_for_recipe(r.adj_config, cfg_msg, hdr="\n\tEdits:")
        print('\n')
    print('\n')


def print_cfg_for_recipe(cfg, fmt, hdr=None):
    if hdr != None:
        print(hdr)

    for section in cfg.keys():
        for item, value in cfg[section].items():
            print(fmt.format(section, item, value))


def print_details(details, mcfg):
    """
    Prints out the details for a list of provided options
    """

    msg = "{: <20} {: <25} {: <25} {: <25} {: <60}"
    hdr  = '\n' + msg.format('Section', 'Item', 'Default', 'Options',
                           'Description')
    print(hdr)
    print('='*len(hdr))
    nopts = len(details)
    # At least a section was provided
    if nopts >= 1:
        if details[0] in mcfg.keys():
            # A section and item was provided
            if nopts == 2:
                if details[1] in mcfg[details[0]].keys():
                    print(msg.format(details[0], details[1],
                                    mcfg[details[0]][details[1]].default,
                                    mcfg[details[0]][details[1]].options,
                                    mcfg[details[0]][details[1]].description))
                else:
                    print("Item {0} in not a registered item."
                          "".format(details[1]))
                    sys.exit()

            # Print the whole section
            else:
                for k, v in mcfg[details[0]].items():
                    print(msg.format(details[0], k,
                                    v.default,
                                    v.options,
                                    v.description))

        # Section does not exist
        else:
            print("Section {0} in not a valid section.".format(details[0]))
            sys.exit()

    else:
        print("Please provide at least a section for information")
        sys.exit()
